/*
	* Copyright (c) 2021-11 by ELECTRONIC ASSEMBLY <technik@lcd-module.de>
	*
	* SmallProtcoll implementation ESP32
	*
	* This file is free software; you can redistribute it and/or modify
	* it under the terms of either the GNU General Public License version 2
	* or the GNU Lesser General Public License version 2.1, both as
	* published by the Free Software Foundation.
*/

#include <Arduino.h>
#include <HardwareSerial.h>

#include "../../debug.h"
#include "../../uniTFTDesignerDefines.h"
#include "uniTFT.h"



//Pin definition EA 94998-E
#define RXD2 12
#define TXD2 13

//Pin definition HardwareSPI ESP32-WROOM, SPI not deeeply tested
#define SPI_CS 15
#define SPI_MOSI 13
#define SPI_MISO 12
#define SPI_CLK 14

//create uniTFT class, allocate the needed space
unitft uniTFT;


//----------------------------------------------------public Functions----------------------------------------------------
//Please use these functions in your sketch
/*----------------------------
	Func: initialize_rs232
	Desc: Initializes RS232 Hardware and sets up connection to uniTFTs
	Vars: baudrate and sbufpin
	Ret : ---
------------------------------*/
void unitft::initialize_rs232 (int baud, byte p_sbuf)
{
	char strbuf[14];
	uniSerial = new HardwareSerial(2);
	initialize(p_sbuf); //General initialisation: Pin, Buffers and so on
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("Init RS232");
	#endif
	uniSerial->setRxBufferSize(2048); //maximum protocol size
	uniSerial->begin(baud, SERIAL_8N1, RXD2, TXD2); //open RS232 port
	interface = INTERFACE_RS232; //remeber which interface is used to communicate
	
	checkReceiveBufferStatus(); //get protocol information
	
	sprintf(strbuf, "#VRF%d,%0.2f\x0a",R_ESPVERSION, VERSION);
	uniTFT.sendDataPacket(strbuf,strlen(strbuf));
	
}

/*----------------------------
	Func: initialize_rs232_newBaud
	Desc: Initializes RS232 Hardware and sets up connection to uniTFTs
	Vars: new baudrate
	Ret : ---
------------------------------*/
void unitft::initialize_rs232_newBaud(int baud)
{
	char strbuf[14];
	if(interface == INTERFACE_RS232 && uniSerial->baudRate() != baud)
	{
		int repeat = 0;
		sprintf(strbuf, "#XCR %d\x0a",baud); //Send module command to change to new baudrate
		sendDataPacket(strbuf,strlen(strbuf), 0);

		#if (DBG_ON && DBG_PROTOCOL)
			Serial.printf("Changing baudrate. need %d repeatcyles to set new baud\r\n", repeat);
		#endif
		uniSerial->flush(); //clear serial buffer
		uniSerial->end(); //end serial to change baudrate
		delay(1);
		uniSerial->setRxBufferSize(2048); //set receive buffer size
		uniSerial->begin(baud, SERIAL_8N1, RXD2, TXD2); //open RS232 port with new baudrate
	}
}

/*----------------------------
	Func: initialize_spi
	Desc: Initializes SPI Hardware and sets up connection to uniTFTs
	Vars: SPI settings and sbufpin
	Ret : ---
------------------------------*/
void unitft::initialize_spi(SPISettings settings, byte p_sbuf)
{
	initialize(p_sbuf); //General initialisation: Pin, Buffers and so on
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("Init SPI\r\n");
	#endif
	//Initialize SPI Hardware
	digitalWrite(SPI_CS, HIGH);
	pinMode(SPI_CS, OUTPUT);
	pinMode(SPI_MOSI, OUTPUT);
	pinMode(SPI_MISO, INPUT);
	pinMode(SPI_CLK, OUTPUT);
	V_SPI = new SPIClass(HSPI);
	V_SPI->begin();
	V_SPI->beginTransaction(settings);
	interface = INTERFACE_SPI; //remeber which interface is used to communicate
    delay(10);
	checkReceiveBufferStatus(); //get protocol information 
}


/*----------------------------
	Func: sendDataPacket
	Desc: sends out a data packet, wrapped in protocol frames
	Vars: *data, length of data
	Ret : transfer packet ok =true, error=false
------------------------------*/
boolean unitft::sendDataPacket(const char *dat, unsigned int len)
{
	byte translen;
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("Send DataPacket: %s\r\n", dat);
	#endif
	while(len)
	{
		translen = (len>252)? 252 : len; //maximum length of SmallProtokoll
		
		while(receiveBufferFree<translen) //receiveBufferFree variable holds free buffersize of module. If not enough space
		{
			checkReceiveBufferStatus(); //get new receiveBufferFree from module
			if(receiveBufferFree<translen)//Check if enough space.
				delay(TIMEOUTBUFFERFULL); //not enough space: Give module a little time to execute
		}
		
		if(sendPacket(dat,translen,DC1) == ACK) //send protocol packet and check if returnvalue is ACK (0x06)
		{
			len -= translen; //recalculate len to send
			receiveBufferFree -= translen;
			dat +=translen;
			#if (DBG_ON && DBG_PROTOCOL)
				Serial.println("Send DataPacket: ok\r\n");
			#endif
		}
		else //Error sending packet
		{
			#if (DBG_ON && DBG_PROTOCOL)
				Serial.println("Send DataPacket: ok\r\n");
			#endif
			return false;
		}
	}
	return true; //data successfully sent. Retrun true for ok
}

/*----------------------------
	Func: check_sendBuffer
	Desc: Receives data from module's send buffer and stores it
	Vars: ---
	Ret : ---
------------------------------*/
void unitft::check_sendBuffer()
{
	if(digitalRead(p_sbuf)==0) //Check SBUF pin. If low, data is present in display's send buffer
	{
		#if (DBG_ON && DBG_PROTOCOL)
			Serial.println("Check Sendbuffer");
		#endif
		if(sendPacket("S", 1, DC2) == ACK) //Protocol command to pull display's send buffer
		{
			int len = recPacket(); //get the returned values and check consitency (Checksum)
			#if (DBG_ON && DBG_PROTOCOL)
				Serial.printf("Receive bytes: %d\r\n", len);
			#endif
			const byte *rec = (const byte*)receivepacket; //pointer to buffer
			while(len--) //copy data to buffer.
			{
				if(rec_nWrite < RECBUFSIZE) 
					recBuf[rec_nWrite++] = *rec++;
				else
				{
					rec_nWrite = 0;
					recBuf[rec_nWrite++] = *rec++;
				}
			} 
		}
	}
}

/*----------------------------
	Func: dataAvailable
	Desc: Returns the count of already received data bytes
	Vars: ---
	Ret : count of received bytes
------------------------------*/
int unitft::dataAvailable()
{
	//Calculate received bytes from ring buffer
	if(rec_nWrite >= rec_nRead)
		return rec_nWrite - rec_nRead;
	else
		return RECBUFSIZE - rec_nRead + rec_nWrite;
}

/*----------------------------
	Func: getByte
	Desc: Returns one byte from receive data bytes
	Vars: ---
	Ret : byte
------------------------------*/
byte unitft::getByte()
{
	byte ret = 0;
	if(dataAvailable() == 0) //no byte in buffer, poll one for one second, otherwise return 0
	{
		unsigned long milis_wait = millis() + 1000;
		while(dataAvailable() == 0 && milis_wait >= millis())
			uniTFT.check_sendBuffer();
	}
	
	if(dataAvailable()) //byte in buffer?
	{
		ret = recBuf[rec_nRead++]; 
		if(rec_nRead>=RECBUFSIZE)
			rec_nRead=0;
	}
	return ret;
}

/*----------------------------
	Func: getString ASCII
	Desc: Returns a character string from receive data bytes
	Vars: buffer to store string bytes, maximum length of stringbuffer in byte
	Ret : ---
------------------------------*/
void unitft::getString(char *str, int maxlen)
{
	char b = 1;
	while(b != 0 && maxlen--) //as long as string is not ended (==0) or maximum is reached
	{
		b = (char)getByte(); //get one byte
		*str++ = b; //store it to buffer
	}		  
}

/*----------------------------
	Func: getString UNICODFE
	Desc: Returns a character string from receive data bytes
	Vars: buffer to store string bytes, maximum length of stringbuffer in 16 bit
	Ret : ---
------------------------------*/
void unitft::getString(char16_t *str, int maxlen)
{
	char16_t b = 1;
	while(b != 0 && maxlen--) //as long as string is not ended (==0) or maximum is reached
	{
		b = (char16_t)getByte();
		b |= (((char16_t)getByte())<<8); //get 16 bit character
		*str++ = b; //and store to buffer
	}			  
}

/*----------------------------
	Func: resetReceiveBuf
	Desc: Clears receive buffer bytes
	Vars: ---
	Ret : ---
------------------------------*/
void unitft::resetReceiveBuf()
{
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("Reset Rx Buffer");
	#endif	
	if(interface == INTERFACE_RS232)
		uniSerial->flush();
	rec_nRead = 0;
	rec_nWrite = 0;
}

/*----------------------------
	Func: rs232_direct_available
	Desc: Direct access to serial port, no protocol routines. Needed for direct connection to bluetooth or wifi
	Vars: ---
	Ret : count of received bytes
------------------------------*/
int unitft::rs232_direct_available()
{
	if(interface == INTERFACE_RS232)
		return uniSerial->available(); //use Arduino function if it's RS232 interface
	
	return 0; //otherwise return 0
}

/*----------------------------
	Func: rs232_direct_read
	Desc: Reads avaible bytes to buffer
	Vars: pointer to buffer, length of buffer
	Ret : count of stored bytes
------------------------------*/
int unitft::rs232_direct_read(byte *buf, int maxlen)
{
	if(interface == INTERFACE_RS232)
		return uniSerial->read(buf, maxlen); //use Arduino function if it's RS232 interface
	
	return 0; //otherwise return 0
}

/*----------------------------
	Func: rs232_direct_read
	Desc: Reads avaible bytes to buffer
	Vars: ---
	Ret : read byte
------------------------------*/
byte unitft::rs232_direct_read()
{
	if(interface == INTERFACE_RS232)
		return uniSerial->read(); //use Arduino function if it's RS232 interface
		
	return 0; //otherwise return 0
}

/*----------------------------
	Func: rs232_direct_write
	Desc: Sends one byte to the RS32 interface
	Vars: byte to send
	Ret : true if sending succeeds
------------------------------*/
boolean unitft::rs232_direct_write(char dat)
{
	if(interface == INTERFACE_RS232)
	{
		uniSerial->write(dat);
		return true;
	}
	
	return false;
}

//----------------------------------------------------private Functions----------------------------------------------------
//normally you don't need those functions in your sketch

/*----------------------------
	Func: initialize
	Desc: Initialize Protocol
	Vars: ---
	Ret : ---
------------------------------*/
void unitft::initialize(byte p_sbuf)
{
	rec_nRead = 0; //buffer init
	rec_nWrite = 0;
	unitft::p_sbuf = p_sbuf; //remember SBUF-Pin port
	
	pinMode(p_sbuf, INPUT_PULLUP); //SBUF-Pin input pull up
	sendBufferReady = 0;
	receiveBufferFree = 0;
}

/*----------------------------
	Func: checkReceiveBufferStatus
	Desc: Checks the Protocol status (DC2 I)
	Vars: ---
	Ret : ---
------------------------------*/
void unitft::checkReceiveBufferStatus()
{
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("Check Receive Buffer Status\r\n");
	#endif
	if(sendPacket("I", 1, DC2) == ACK) //Protcol status request
	{
		if(recPacket() == 2) //decode returned bytes
		{
			sendBufferReady = receivepacket[0]; //bytes ready to send from module to master
			receiveBufferFree = receivepacket[1]; //buffer bytes free in module
			#if (DBG_ON && DBG_PROTOCOL)
				Serial.println("Check Receive Buffer Status ok\r\n");
			#endif
		}
	}
}

/*----------------------------
	Func: sendDataPacket
	Desc: sends out a data packet, wrapped in protocol frames. Decide repetitions
	Vars: *data, length of data
	Ret : transfer packet ok =true, error=false
------------------------------*/
boolean unitft::sendDataPacket(const char *dat, unsigned int len, int rep)
{
	byte translen;
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("Send DataPacket: %s\r\n", dat);
	#endif
	while(len)
	{
		translen = (len>252)? 252 : len; //maximum length of SmallProtokoll
		
		while(receiveBufferFree<translen) //receiveBufferFree variable holds free buffersize of module. If not enough space
		{
			checkReceiveBufferStatus(); //get new receiveBufferFree from module
			if(receiveBufferFree<translen)//Check if enough space.
				delay(TIMEOUTBUFFERFULL); //not enough space: Give module a little time to execute
		}
		
		if(sendPacket(dat,translen,DC1, rep) == ACK) //send protocol packet and check if returnvalue is ACK (0x06)
		{
			len -= translen; //recalculate len to send
			receiveBufferFree -= translen;
			dat +=translen;
			#if (DBG_ON && DBG_PROTOCOL)
				Serial.println("Send DataPacket: ok\r\n");
			#endif
		}
		else //Error sending packet
		{
			#if (DBG_ON && DBG_PROTOCOL)
				Serial.println("Send DataPacket: ok\r\n");
			#endif
			return false;
		}
	}
	return true; //data successfully sent. Retrun true for ok
}

/*----------------------------
	Func: sendPacket
	Desc: sends out a protocol packet
	Vars: pointer to data, length of data, type (DC1 ->send data or DC2 -> send protocol command), maximum repitions
	Ret : Ack-Byte (0x06 = ok, 0x15 = NACK, other value no answer)
------------------------------*/
byte unitft::sendPacket(const char *dat, byte len, byte type, int repmax)
{
	if(len>=252)
		return 0;
	
	byte *p_send = sendpacket;
	byte sendlen = len;
	byte bcc = type;
	*p_send++ = type;
	*p_send++ = sendlen;
	bcc += sendlen;
	while(len--) //calculate checksum and construct sendbuffer
	{
		*p_send++ = *dat;
		bcc += *dat++;
	}
	*p_send = bcc;
	
	byte ack;
	byte rep = 0;
	
	
	do //send data to module and receive ack. Possibly repeat packet if not recevied
	{
		#if (DBG_ON && DBG_PROTOCOL)
			Serial.printf("Send Packet: %s\r\n", sendpacket);
		#endif
		out(sendpacket, sendlen+3); //send data
		delayMicroseconds(6); //wait till module is ready to answer
		#if (DBG_ON && DBG_PROTOCOL)
			Serial.println("Want to rec ACK");
		#endif
		ack = rec(); //rec ack-byte
		rep++; //increment repeats
		if(ack != ACK && ack!= NACK && rep < SENDREPEAT) //If returnvalue is not ACK or NACK, wait TIMEOUT (see uniTFT.h)
			delay(TIMEOUT);
		else if (ack == NACK && rep < SENDREPEAT) //if returnvalue is NACK. Wait as module's receivebuffer seems to be full
			delay(TIMEOUTBUFFERFULL);
	}while(ack != ACK && rep < repmax); //reapeat if answer is not ACK and repetitions are smaller than SENDREPEAT (see uniTFT.h)
	
	return ack;
}

/*----------------------------
	Func: recPacket
	Desc: receives data and writes to global receive buffer, return length of valid data
	Vars: ---
	Ret : count of received bytes from uniTFT module
------------------------------*/
byte unitft::recPacket()
{
	byte *p_dat = receivepacket;
	
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("Rec packet");
	#endif
	
	if(rec_buf(receivepacket, 2) == false) //error receiving two bytes
	{
		receivepacket[0] = 0;
		receivepacket[1] = 0;
	}
    byte type = receivepacket[0]; //First byte is type (DC1 or DC2)
	byte len = receivepacket[1]; //len of data bytes
	byte counter = len;
	byte bcc = type + len; //bcc calculation
	
	rec_buf(receivepacket, counter); //receive data bytes
	while(counter--) //calculate checksum
		bcc += *p_dat++; 
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("Rec. %d bytes. Want to rec BCC\r\n", len);
	#endif
	if(bcc == rec()) //if calculated bcc equals uniTFT module's bcc
		return len; //return length
	else
		return 0; //otherwise retun length 0
}

/*----------------------------
	Func: out
	Desc: wrapper for sending
	Vars: buffer, len
	Ret : ---
------------------------------*/
void unitft::out(byte *dat, byte len)
{
	if(interface == INTERFACE_RS232) //send through RS232
		out_rs232(dat, len);
	else if(interface == INTERFACE_SPI) //send through SPI (untested)
		out_spi(dat, len);
}

/*----------------------------
	Func: rec
	Desc: wrapper for receiving
	Vars: ---
	Ret : received byte
------------------------------*/
byte unitft::rec()
{
	if(interface == INTERFACE_RS232) //receive through RS232
		return rec_rs232();
	else if(interface == INTERFACE_SPI) //receive through SPI (untested), sending 0xFF as dummy bytes
		return rec_spi();
	
	return 0;
}

/*----------------------------
	Func: rec_buf
	Desc: wrapper for receiving multible bytes
	Vars: *buffer to store data, length
	Ret : true on success, false on error
------------------------------*/
boolean unitft::rec_buf(byte *buf, int len)
{
	if(interface == INTERFACE_RS232) //receive through RS232
		return rec_buf_rs232(buf, len);
	else if(interface == INTERFACE_SPI) //receive through SPI (untested), sending 0xFF as dummy bytes
		return rec_buf_spi(buf, len);
	
	return false;
}


/*----------------------------
	Func: out_rs232
	Desc: Sends a buffer with length, blocks till all bytes are sent
	Vars: buffer, len
	Ret : ---
------------------------------*/
void unitft::out_rs232(byte *dat, byte len)
{ 
	int sendlen;
	while(len)
	{
		sendlen = len;
		if(len>32) //maximum buffer size is 32 in Arduino serial port
			sendlen = 32;
		len-=sendlen;
		
		uniSerial->write(dat, sendlen); 
		dat+=sendlen;
	}
	uniSerial->flush(); //wait till data is sent
}

/*----------------------------
	Func: rec_rs232
	Desc: receives one byte
	Vars: ---
	Ret : received byte
------------------------------*/
byte unitft::rec_rs232() 
{
	byte rec = 0x16; //dummy value for error
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("Want to rec 1 Byte");
	#endif
	unsigned long myTime = millis()+500; //timeout 500 ms for no byte present in serial receive buffer
	
	while(!uniSerial->available() && myTime!=millis())
		;
	
    if(uniSerial->available())
		rec = uniSerial->read();
	
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("received 0x%02X=%c\r\n", rec);
	#endif
	return rec;
}

/*----------------------------
	Func: rec_buf_rs232
	Desc: receives multible bytes
	Vars: *buffer to store data, length
	Ret : true on success, false on error
------------------------------*/
boolean unitft::rec_buf_rs232(byte *buf, int len)
{
	unsigned long myTime;
	
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("Want to rec %d bytes\r\n", len);
	#endif
	while(len)
	{  	
		myTime = millis()+500;
		
		while(!uniSerial->available() && myTime!=millis()) //timeout 500 ms for no byte present in serial receive buffer
			;
		if(uniSerial->available()==0)
		{
			break;
		}
		else
		{
			
			while(uniSerial->available() && len) //as long as data is available and data needs to be received
			{
				len--;
				*buf++ = uniSerial->read(); //store received byte to buffer
				#if (DBG_ON && DBG_PROTOCOL)
					Serial.printf("0x%02X=%c, ", *(buf-1), *(buf-1));
				#endif
			}
		}
	}
	
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("\r\nRec ended");
	#endif
	return (len==0);
}


/*----------------------------
	Func: spi_out
	Desc: Sends a buffer with length
	Vars: buffer, len
	Ret : ---
------------------------------*/
void unitft::out_spi(byte *dat, byte len)
{ 
	int sendlen;
	digitalWrite(SPI_CS, LOW); //CS low
	while(len)
	{
		sendlen = len;
		if(len>32) //Sendbuffer is 32 for smaller Arduinos
			sendlen = 32;
		len-=sendlen;
		
		V_SPI->transfer(dat, sendlen); //transfer bytes
		dat+=sendlen;
	}
	digitalWrite(SPI_CS, HIGH); //CS high
}

/*----------------------------
	Func: spi_rec
	Desc: Sends 0xFF and receives one
	Vars: ---
	Ret : received byte
------------------------------*/
byte unitft::rec_spi() 
{
	byte rec = 0x16; //Dummy byte for error
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("SPI: Want to rec 1 Byte");
	#endif
	digitalWrite(SPI_CS, LOW);
	rec = V_SPI->transfer(0xFF); //receive one byte, send dummy byte 0xFF
	digitalWrite(SPI_CS, HIGH);
	
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("SPI: received 0x%02X=%c\r\n", rec);
	#endif
	return rec;
}

/*----------------------------
	Func: spi_rec
	Desc: receives multible bytes and send 0xFF
	Vars: *buffer to store data, length
	Ret : true (function returns boolean for compatibility with RS232)
------------------------------*/
boolean unitft::rec_buf_spi(byte *buf, int len)
{
	
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.printf("SPI: Want to rec %d bytes\r\n", len);
	#endif
	digitalWrite(SPI_CS, LOW);
	
	while(len--)
	{
		*buf++ = V_SPI->transfer(0xFF); //receive one byte, send dummy byte 0xFF and store it to buffer
		#if (DBG_ON && DBG_PROTOCOL)
			Serial.printf("0x%02X=%d, ", *(buf-1), *(buf-1));
		#endif
	}
	
	digitalWrite(SPI_CS, HIGH);
	#if (DBG_ON && DBG_PROTOCOL)
		Serial.println("\r\nSPI Rec ended");
	#endif
	return true;
}

